
/* snoscan  - methylation guide snoRNA search program
 *
 * by Todd MJ Lowe    v. 0.1       12/08/96
 *                    v. 0.9 beta  02/19/99
 *                    v. 0.9.1     04/26/17 - updates squid library (SRE)
 *
 * copyright 1999 by T. Lowe
 *
 * Uses Sean Eddy's function library for biological sequence analysis
 * (Squid v1.5.11)
 *
 *  matrices.h  - matrices used for scoring snoRNA features
 */

/* These matrices were generated from a collection of Perl scripts,
 *  using feature alignments, or collections of annotated snoRNA
 *  sequences. 
 *  
 * On my list of things ToDo: combine these scripts into one single
 * program that takes a snoRNA alignment as input.  These
 * yeast-trained matrices actually work quite well for all euks i've
 * scanned, so these defaults should work fine for general use.
 *
 * (A special set of matrices were developed to scan archaeal genomes
 * for snoRNAs, but these are still experimental and aren't ready for
 * general release to the public.)
 */ 
 
/* To see how all these matrices fit together in the overall scoring
 * scheme, see the function ScoreHit() in search.c   
 *
 * Some scoring values are still embeded in functions in search.c
 * (egad! what bad style!) -- I will move those over into this header
 * file next major work I do on the code 
 */


/* Yeast C and D' Box matrices 1/25/98 
   trained on 41 known/exp confirmed snos
   (used "gen-matrix.pl" Perl script with
   hand-alignment of feature seqs
*/

float Sc2_Cbox_Mat[5][CBOX_LEN] = {
  {1.145,-2.492,-2.492,1.171,-2.492,-2.492,0.973},
  {-2.003,-2.003,-2.003,-2.003,-2.003,-2.003,-2.003},
  {-2.003,1.608,-2.003,-2.003,1.661,-2.003,-0.393},
  {-2.492,-1.799,1.119,-2.492,-2.492,1.145,-1.799},
  {-3.135,-3.135,-3.135,-3.135,-3.135,-3.135,-3.135}
};

float Sc2_DpBox_Mat[5][DBOX_LEN] = {
  {1.094,-2.125,-1.432,-1.026},
  {-1.635,-1.635,-0.536,1.409},
  {-1.635,1.661,-1.635,-1.635},
  {-2.125,-2.125,0.920,-1.026},
  {-3.135,-3.135,-3.135,-3.135}
};

/* rRNA Complementarity Matrix 
   01/25/98 
   generated with "parse-snos.pl" Perl script
   from all 41 known yeast snoRNAs

   {A-A,A-C,A-G,A-U,A-N}
   {C-A,... 
   where first base is rRNA
   and second from query 
   (i.e. C-A  C (rRNA) A (query) */

float Sc2_Compl_Mat[4][5] = {
  {-1.898,-1.409,-1.409,1.137,-1.30},
  {-1.626,-1.137,1.615,-1.626,-1.30},
  {-1.689,1.552,-1.200,-1.401,-1.30},
  {1.088,-1.457,-1.170,-1.947,-1.30}
};


/* Terminal stem matrix
   generated with "term-stem-matrix" Perl script

   {A-A,A-C,A-G,A-U,A-N}
   {C-A,...                */ 

float TermStem_Mat[5][5] = {
  {-1.324,-0.834,-0.834,0.884,-1.3},
  {-1.324,-0.834,1.373,-1.324,-1.3},
  {-0.908,1.016,-0.419,-0.320,-1.3},
  {0.527,-0.419,0.169,-0.908,-1.3},
  {-1.3,-1.3,-1.3,-1.3,-1.3}
};

/* Score matrices for length distributions of 
   snoRNA-rRNA duplex region lengths
   -matrices generated by arbitrarily defining bins length ranges,
   then hand-calculating scores based on bin contents

   Scores in Nats ln([counts in bin]/[totalcounts in all bins])
   Re-trained on 41 snos (50 compl regions) 10/98 */

#define GUIDE_DUPLEX_LEN_MAT_SIZE 6

int GuideDuplexLenBins_Mat [GUIDE_DUPLEX_LEN_MAT_SIZE] = {9,12,15,18,21,1000};

float GuideDuplexLenSc_Mat [GUIDE_DUPLEX_LEN_MAT_SIZE] = 
{-3.91,-0.776,-1.02,-2.30,-2.81,-3.91};
  

/* Score matrices for length distributions of gaps between features
   -again, hand-calculated
   Scores are in Nats ln([counts in bin]/[totalcounts in all bins])
   Trained from 41 Sc snoRNAs 1/26/98 
*/

/* Matrix scoring distance between C Box and rRNA compl region
   for D' box containing snoRNAs */

#define CBOX_COMPLDp_MAT_SIZE 6

/* bin range delimiters */
int CBxComplDp_DistIdx_Mat[CBOX_COMPLDp_MAT_SIZE] = {5,10,15,20,25,1000};
/* scores in bins */
float CBxComplDp_DistSc_Mat[CBOX_COMPLDp_MAT_SIZE] = 
{-3.30,-1.10,-1.50,-1.22,-2.19,-5.60};  

/* Matrix scoring distance between compl region and D Box 
   for D' box containing snoRNAs */

#define COMPL_DBOX_MAT_SIZE 7

/* bin range delimiters */
int ComplDBx_DistIdx_Mat[COMPL_DBOX_MAT_SIZE] = {25,35,45,55,65,90,1000};
/* scores in bins */
float ComplDBx_DistSc_Mat[COMPL_DBOX_MAT_SIZE] =
{-5.60,-1.91,-1.10,-1.69,-1.50,-2.19,-5.60};

/* Matrix scoring distance between C Box and rRNA compl region
   for non-D' box containing snoRNAs */

#define CBOX_COMPL_MAT_SIZE 6

/* bin range delimiters */
int  CBxCompl_DistIdx_Mat[CBOX_COMPL_MAT_SIZE] = {45,55,65,75,130,1000};
/* scores in bins */
float CBxCompl_DistSc_Mat[CBOX_COMPL_MAT_SIZE] = {-2.83,-1.45,-1.04,-1.04,-2.83,-5.13};   


/* Scores used for two types of D boxes found:
   (CUGA or AUGA)  

   updated   1/26/98

(C/A)UGA   

      U and A  log(39/39 /0.31) = 1.17 
      G        log(39/39 /0.19) = 1.66
      C or A   C occurs 36/39 times log(36/39 / 0.19) =  1.58
               A occurs  3/39 times log( 3/39 / 0.31) = -1.39  

Total score for CUGA = 1.17*2 + 1.66 + 1.58 = 5.58 nats (8.03 bits)
                AUGA = 1.17*2 + 1.66 - 1.39 = 2.61 nats (3.76 bits)
*/

#define D_BOX_CUGA_SCORE 5.58
#define D_BOX_AUGA_SCORE 2.61

